/*  fwdmodel_IR.cc - Implements T1 estimation using the Variable Flip Angle approach

    Alex Smith, FMRIB

    Copyright (C) 2007-2017 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#include "fwdmodel_IR.h"

#include "fabber_core/tools.h"

#include "armawrap/newmat.h"

#include <string>
#include <vector>

using namespace NEWMAT;
using namespace std;

FactoryRegistration<FwdModelFactory, IRFwdModel> IRFwdModel::registration("ir");

static OptionSpec OPTIONS[] = {
    { "tis-file", OPT_MATRIX, "File containing a list of inversion times (s)", OPT_REQ, "" },
    { "invefficiency", OPT_BOOL, "If specified, will also calculate the inversion pulse efficiency", OPT_NONREQ, "" },
    { "" }
};

void IRFwdModel::GetOptions(vector<OptionSpec> &opts) const
{
    for (int i = 0; OPTIONS[i].name != ""; i++)
    {
        opts.push_back(OPTIONS[i]);
    }
}

std::string IRFwdModel::GetDescription() const
{
    return "Calculates T1 from inversion recovery data, assuming the signal has fully relaxed between each image.";
}

string IRFwdModel::ModelVersion() const
{
    string version = "fwdmodel_IR.cc";
#ifdef GIT_SHA1
    version += string(" Revision ") + GIT_SHA1;
#endif
#ifdef GIT_DATE
    version += string(" Last commit ") + GIT_DATE;
#endif
    return version;
}

void IRFwdModel::Initialize(FabberRunData &rundata)
{
    string TIs = rundata.GetStringDefault("tis-file", "");
    m_TI = fabber::read_matrix_file(TIs);

    m_InvEff = false;
    m_InvEff = rundata.ReadBool("invefficiency");
}

void IRFwdModel::NameParams(vector<string> &names) const
{
    names.clear();

    names.push_back("T1");
    names.push_back("sig0");
    if (m_InvEff)
        names.push_back("InvEff");
}

void IRFwdModel::HardcodedInitialDists(MVNDist &prior, MVNDist &posterior) const
{
    assert(prior.means.Nrows() == NumParams());

    SymmetricMatrix precisions = IdentityMatrix(NumParams()) * 1e-12;

    int place = 1;
    // Set priors
    // T1
    prior.means(place) = 1;
    precisions(place, place) = 0.44; // Based off of similar prior precision in FABBER model
    ++place;

    // S0
    prior.means(place) = 0.01;
    precisions(place, place) = 1e-12;
    ++place;

    if (m_InvEff)
    {
        prior.means(place) = 1;
        precisions(place, place) = 1e99;
        ++place;
    }

    // Set precsions on priors
    prior.SetPrecisions(precisions);

    // Set initial posterior
    posterior = prior;

    // For parameters with uniformative prior choose more sensible inital posterior
    // Tissue perfusion
    // posterior.means(T1_index()) = 1;
    // precisions(T1_index(), T1_index()) = 0.1;

    // posterior.SetPrecisions(precisions);
}

void IRFwdModel::Evaluate(const ColumnVector &params, ColumnVector &result) const
{
    // ensure that values are reasonable
    // negative check
    ColumnVector paramcpy = params;
    for (int i = 1; i <= NumParams(); i++)
    {
        if (params(i) < 0)
        {
            paramcpy(i) = 0;
        }
    }

    // parameters that are inferred - extract and give sensible names
    double T1;
    double sig0;   //'inital' value of the signal
    double InvEff; // B1 Correction Factor
    int place = 1;

    // extract values from params
    // T1
    T1 = paramcpy(place);
    ++place;

    sig0 = paramcpy(place);
    ++place;

    if (m_InvEff)
    {
        InvEff = paramcpy(place);
        ++place;
    }
    else
    {
        InvEff = 1;
    }

    // Keep sensible values for parameters
    if (T1 < 1e-8)
        T1 = 1e-8;
    if (sig0 < 1e-8)
        sig0 = 1e-8;

    int ntpts = m_TI.Nrows();
    if (data.Nrows() != ntpts)
    {
        throw FabberRunDataError("Number of volumes in data does not match number of inversion times");
    }

    // --- IR Function ----
    ColumnVector sig(ntpts);
    sig = 0.0;
    for (int i = 1; i <= ntpts; i++)
    {
        sig(i) = fabs(sig0 * (1 - 2 * InvEff * exp(-m_TI(i) / T1)));
    }
    result = sig;

    for (int i = 1; i <= ntpts; i++)
    {
        if (isnan(result(i)) || isinf(result(i)))
        {
            exit(0);
            LOG << "Warning NaN of inf in result" << endl;
            LOG << "result: " << result.t() << endl;
            LOG << "params: " << params.t() << endl;

            result = 0.0;
            break;
        }
    }
}

FwdModel *IRFwdModel::NewInstance()
{
    return new IRFwdModel();
}
